"""
    This module provides the communication protocol seen from the point
    of view of the garbler. First, the garbler and the evaluator establish
    a connection through a socket. Then the garbler creates the circuit
    and garbles all the gates. He then sends to the evaluator
    the wire identifiers so that the evaluator can choose which truth
    values to supply to each wire he controls. After this,
    the input labels are transferred to the evaluator.
    Following the garbled circuits protocol, the garbler's labels can be
    sent *as is*, as they are obfuscated so the evaluator can not learn
    anything. The evaluator's labels however are trickier,
    so a 1-out-of-2 oblivious transfer protocol must be followed
    for each input label the evaluator supplies.

    Once the evaluator is in possesion of all the input labels,
    he can reconstruct the circuit and send the final output label
    to the garbler. The garbler can then compare the label in his
    circuit and decide which truth value it corresponds.
    Finally, the garbler sends the evaluator the final truth value.

"""

import copy
import gabes.network as net

from gabes.circuit import Circuit
from gabes.utils import ask_for_inputs
from gabes.ot import garbler_ot


def garbler(args):
    """
        The main function of the application for the garbler. For
        more information on the process, see above.


        :param args: the arguments from the command line interface
        :return: the output of the circuit
        :rtype: bool

    """
    print("Welcome, garbler. Waiting for the evaluator...")
    sock, client = net.connect_garbler(args.address)
    circ = Circuit(args.circuit)
    print("Circuit created...")
    identifiers = hand_over_wire_identifiers(client, circ)
    if args.identifiers:
        inputs = {x: y for x, y in zip(args.identifiers, args.bits)}
    else:
        inputs = ask_for_inputs(identifiers)
    hand_over_labels(client, circ, inputs)
    hand_over_cleaned_circuit(client, circ)
    final_output = learn_output(client, circ)
    print("The final output of the circuit is: {}".format(final_output))
    sock.close()
    return final_output


def hand_over_wire_identifiers(client, circ):
    """
        Sends the wire identifiers to the evaluator.

        :param client: the client is the evaluator
        :param circ: the circuit to which the wires belong
        :return: the identifiers of the input wires
        :rtype: list(str)

    """
    identifiers = [wire.identifier for wire in circ.get_input_wires()]
    net.send_data(client, identifiers)
    net.wait_for_ack(client)
    return identifiers


def hand_over_cleaned_circuit(client, circ):
    """
        Sends a clean circuit (in which every label's *represents*
        flag has been deleted) to the evaluator.

        :param client: the client is the evaluator
        :param circ: the circuit in question

    """
    new_circ = circ.clean()
    net.send_data(client, new_circ)


def hand_over_labels(client, circ, garbler_inputs):
    """
        Sends the input labels of the circuit to the evaluator. The labels
        that belong to the garbler can be sent without any modification.
        In order for the evaluator to learn his labels, he must acquire
        them through the oblivious transfer protocol, in which the
        garbler inputs the two possible labels, the evaluator inputs
        his choice of truth value, and the evaluator learns which
        label corresponds to his truth value without the garbler learning
        his choice and without the evaluator learning both labels.

        :param client: the client is the evaluator
        :param circ: the circuit in question
        :param garbler_inputs: the inputs the garbler provides

    """
    identifiers = set(wire.identifier for wire in circ.get_input_wires())
    if not set(garbler_inputs.keys()).issubset(set(identifiers)):
        raise ValueError('You have supplied a wire '
                         'identifier not in the circuit.')
    for wire in circ.get_input_wires():
        if wire.identifier in garbler_inputs:
            chosen_bit = garbler_inputs[wire.identifier]
            if chosen_bit == '1':
                secret_label = wire.true_label
            else:
                secret_label = wire.false_label
            net.send_data(client, secret_label)
            net.wait_for_ack(client)
        else:
            false_label = copy.deepcopy(wire.false_label)
            true_label = copy.deepcopy(wire.true_label)
            # Clean before sending
            false_label.represents = true_label.represents = None
            garbler_ot(client, false_label, true_label)


def learn_output(client, circ):
    """
        Learns the final truth value of the circuit by comparing the label that
        was handed to him by the evaluator to the two labels in the root of the
        tree (i.e. the final gate).

        :param client: the client is the evaluator
        :param circ: the circuit in question
        :return: the output of the circuit
        :rtype: bool

    """
    output_label = net.receive_data(client)
    output_gate = circ.tree.name
    out1 = output_gate.output_wire.true_label.to_base64()
    out2 = output_label.to_base64()
    output = out1 == out2
    net.send_data(client, output)
    return output



"""
This module implements the :class:`Wire` object. Each gate will have
a left wire (in the case of input gates, this will be probably be supplied
by the garbler), a right wire (evaluator), and an output wire. Each
wire holds the two possible for labels that run through it.
"""

import random
import gabes.settings as settings

from gabes.utils import xor
from gabes.label import Label


class Wire(object):
    """
        The :class:`Wire` object holds two labels representing
        *True* and *False*. In classical garbled circuits,
        there is no need for a point-and-permute bit. In all the other
        cases, a pp_bit is associated to each label. The two labels
        in the same wire must have opposing pp_bits.

        If the optimization chosen is FreeXOR or Half Gates then
        the true label is the false label xored with the global
        parameter `R` defined in :class:`gabes.circuit.Circuit`

        :param str identifier: (optional) wire's unique identifier

        .. code-block:: python

               >>> from gabes.wire import Wire
               >>> w = Wire(identifier='A')
               >>> str(w)
               'Wire A'
               >>> w.false_label
               b'dnE2Gsvhx84HgwrLRm8L9aFtI_aBYxzEDaOBRK2qkP0='
               >>> w.true_label
               b'eBRWiJzYL65gU8nBFvXRZ8NK4_Cf9GlrYtNGZNEZOSs='
               >>> w.get_label(True) == w.true_label.represents
               True
    """
    def __init__(self, identifier=None):
        self.identifier = identifier
        if settings.CLASSICAL:
            self.false_label = Label(False)
            self.true_label = Label(True)
        else:
            b = random.choice([True, False])
            self.false_label = Label(False, pp_bit=b)
            self.true_label = Label(True, pp_bit=not b)
        if settings.FREE_XOR or settings.HALF_GATES:
            self.true_label.label = xor(self.false_label.label, settings.R)

    def __str__(self):
        if self.identifier:
            return "Wire {}".format(self.identifier)
        else:
            return "Unidentified Wire"

    def labels(self):
        """
            A getter method to get the two labels (*False* and *True*)
            going through the wire.

            :return: a tuple of labels
            :rtype: Generator[:class:`Wire`]

        """
        for label in (self.false_label, self.true_label):
            yield label

    def get_label(self, representing):
        """
            Gets the label according to which truth value it represents.

            :param bool representing: *True* for the true label and *False* \
            for the false label
            :return: the corresponding label
            :rtype: :class:`Label`
        """
        return self.true_label if representing else self.false_label



"""
This module handles global settings of the package.
"""

NUM_BYTES = 32

# Optimizations
CLASSICAL = True
POINT_AND_PERMUTE = False
GRR3 = False
FREE_XOR = False
FLEXOR = False
HALF_GATES = False



"""
    This module includes utility functions used throughout the package.
"""

import gabes.settings as settings
import os


def ask_for_inputs(identifiers):
    """
        CLI helper function that queries the user to indicate which
        identifier he supplies and the his choice for each identifier.

        :param list(str) identifiers: the identifiers of the input wires
        :return: the identifiers the user supplies
        :rtype: dict

    """
    print('To start the protocol please indicate with y/n '
          'which wires you supply:')
    chosen_wires = []
    for identifier in identifiers:
        decision = input("Do you supply {}? ".format(identifier))
        while True:
            if decision == 'y':
                chosen_wires.append(identifier)
                break
            elif decision == 'n':
                break
            else:
                decision = input("Sorry, didn't recognize that. \
                                  Indicate with y or n whether \
                                  you supply {}: ".format(identifier))

    inputs = {}
    for identifier in chosen_wires:
        label_choice = input("Choice for {}: ".format(identifier))
        while True:
            if label_choice in '01':
                break
            else:
                label_choice = input("Sorry, you must supply either a 0 \
                                      (indicating false) or a 1 \
                                      (indicating true) for \
                                      wire {}: ".format(identifier))

        inputs[identifier] = label_choice

    return inputs


def get_last_bit(label):
    """
        Gets the last bit from a bytestring.

        :param bytes label: any bytes object
        :return: the last bit
        :rtype: bool

        .. code-block:: python

            >>> import os
            >>> from gabes.utils import get_last_bit
            >>> b1 = os.urandom(10)
            >>> b1
            b'\\xf3\\x9e\\xb0w,|\\xd9\\xa8\\xd73'
            >>> get_last_bit(b1)
            False

    """
    last_byte = label[-1]
    return bool((last_byte >> 7) & 1)


def xor(b1, b2):
    """
        XORs two bytestrings.

        :param bytes b1: first argument
        :param bytes b2: second argument
        :return: the XORed result
        :rtype: bytes

        .. code-block:: python

            >>> import os
            >>> from gabes.utils import get_last_bit
            >>> b1 = os.urandom(10)
            >>> b2 = os.urandom(10)
            >>> b1, b2
            (b'\\xf8\\x00r\\xaf\\x9a\\x06!68\\x83', b'\\x88
            \\xee\\x1c,a\\xd0^\\x8a\\xb4\\xf2')
            >>> xor(b1, b2)
            b'p\\xeen\\x83\\xfb\\xd6\\x7f\\xbc\\x8cq'
            >>> xor(b1, xor(b1, b2)) == b2
            True

    """
    n1 = int.from_bytes(b1, byteorder='big')
    n2 = int.from_bytes(b2, byteorder='big')
    xored = n1 ^ n2
    return int.to_bytes(xored, length=settings.NUM_BYTES, byteorder='big')


def adjust_wire_offset(wire):
    """
        Adjusts the wire's offset so that the two labels have a
        distinct last bit.

        :param wire: the wire in question

    """
    false_label = get_last_bit(wire.false_label.label)
    true_label = get_last_bit(wire.true_label.label)
    while false_label == true_label:
        wire.true_label.label = os.urandom(settings.NUM_BYTES)
        true_label = get_last_bit(wire.true_label.label)



"""
    This module provides the communication protocol seen from the point
    of view of the evaluator. To learn the whole process,
    see the Garbler's section.
"""

import gabes.network as net

from gabes.ot import evaluator_ot
from gabes.utils import ask_for_inputs


def evaluator(args):
    """
        The main function of the application for the evaluator. For
        more information on the process, see the introduction to
        the Garbler's section.


        :param args: the arguments from the command line interface
        :return: the output of the circuit
        :rtype: bool

    """
    print("Welcome, evaluator. Waiting for the garbler...")
    sock = net.connect_evaluator(args.address)
    idents = request_wire_identifiers(sock)
    if args.identifiers:
        inputs = {x: y for x, y in zip(args.identifiers, args.bits)}
    else:
        inputs = ask_for_inputs(idents)
    labels = request_labels(sock, idents, inputs)
    circ = request_cleaned_circuit(sock)
    print("Reconstructing circuit...")
    secret_output = circ.reconstruct(labels)
    final_output = learn_output(sock, secret_output)
    print("The final output of the circuit is: {}".format(final_output))
    sock.close()
    return final_output


def request_wire_identifiers(sock):
    """
        Receives the wire identifiers from the garbler.

        :param sock: the socket from which it will receive the data
        :return: the identifiers of the input wires
        :rtype: list(str)

    """
    identifiers = net.receive_data(sock)
    net.send_ack(sock)
    return identifiers


def request_cleaned_circuit(sock):
    """
        Receives a clean circuit (in which every label's *represents*
        flag has been deleted) from the garbler.

        :param sock: the socket from which it will receive the data
        :return: the cleaned circuit
        :rtype: :class:`Circuit`

    """
    circuit = net.receive_data(sock)
    return circuit


def request_labels(sock, identifiers, evaluator_inputs):
    """
        Receives the input labels of the circuit from the garbler. The labels
        that belong to the garbler can be sent without any modification.
        In order for the evaluator to learn his labels, he must acquire
        them through the oblivious transfer protocol, in which the
        garbler inputs the two possible labels, the evaluator inputs
        his choice of truth value, and the evaluator learns which
        label corresponds to his truth value without the garbler learning
        his choice and without the evaluator learning both labels.

        :param sock: the socket from which it will receive the data
        :param identifiers: the identifiers for all the input wires
        :param evaluator_inputs: the inputs the evaluator provides
        :return: the input labels
        :rtype: list(:class:`Label`)

    """
    labels = []
    if not set(evaluator_inputs.keys()).issubset(set(identifiers)):
        raise ValueError('You have supplied a wire '
                         'identifier not in the circuit.')
    for identifier in identifiers:
        if identifier in evaluator_inputs:
            chosen_bit = evaluator_inputs[identifier]
            secret_label = evaluator_ot(sock, chosen_bit)
        else:
            secret_label = net.receive_data(sock)
            net.send_ack(sock)
        labels.append(secret_label)
    return labels


def learn_output(sock, secret_output):
    """
        Sends the final label and learns the final truth value
        from the garbler.

        :param sock: the socket from which it will receive the data
        :param secret_output: the final label of the circuit
        :return: the output of the circuit
        :rtype: bool

    """
    net.send_data(sock, secret_output)
    final_output = net.receive_data(sock)
    return final_output



import sys
sys.path.append("/Users/nacho/Desktop/gabes")

import argparse
import gabes.settings as settings
from gabes.garbler import garbler
from gabes.evaluator import evaluator


def sanitize_inputs(parser):
    args = parser.parse_args()
    if not args.garbler and not args.evaluator:
        parser.error('Either -g or -e needs to be set.')
    if args.garbler and args.evaluator:
        parser.error('Can not be both garbler and evaluator!')
    if args.garbler and not args.circuit:
        parser.error('The garbler needs to supply the circuit file.')
    if args.bits and not args.identifiers:
        parser.error('-i must be supplied along -b')
    if args.identifiers and not args.bits:
        parser.error('-b must be supplied along -i')
    if args.identifiers and args.bits:
        if len(args.identifiers) != len(args.bits):
            parser.error('Must supply the same number of bits as identifiers')


def sanitize_optimizations(parser):
    args = parser.parse_args()
    if args.classical:
        if any([args.point_and_permute, args.grr3, args.free_xor,
                args.flexor, args.half_gates]):
            parser.error('Classical garbled circuits is not compatible with \
                         any optimization.')
    if args.point_and_permute:
        if any([args.classical, args.grr3, args.free_xor,
                args.flexor, args.half_gates]):
            parser.error('The -pp flag is for the point-and-permute \
                          technique on its own. All successive optimizations \
                          include point-and-permute.')
    if args.free_xor and args.flexor:
        parser.error('FreeXOR is not compatible with FleXOR.')
    if args.flexor and args.half_gates:
        parser.error('FleXOR is not compatible with half gates.')

    settings.CLASSICAL = args.classical
    settings.POINT_AND_PERMUTE = args.point_and_permute
    settings.GRR3 = args.grr3
    settings.FREE_XOR = args.free_xor
    settings.FLEXOR = args.flexor
    settings.HALF_GATES = args.half_gates


def cli():
    parser = argparse.ArgumentParser(description='Program to garble and \
                                     evaluate a circuit.',
                                     epilog='Example usage: gabes -g -c \
                                     Desktop/test.circuit -a 192.08.12.33:1932\
                                      -i A B H -b 001')

    # General options
    parser.add_argument('-g', '--garbler', action='store_true',
                        help="Set this flag to become the garbler")
    parser.add_argument('-e', '--evaluator', action='store_true',
                        help="Set this flag to become the evaluator")
    parser.add_argument('-b', '--bits', metavar="bits",
                        help="Include your private input bitstring \
                              to the circuit (e.g. 001011)")
    parser.add_argument('-i', '--identifiers', nargs='+',
                        metavar="identifier",
                        help="Indicate which input wires you supply to \
                        the circuit (e.g. -i A C D)")
    parser.add_argument('-c', '--circuit',
                        metavar="file",
                        help="Path of the file representing the circuit. \
                              Only the garbler needs to supply the file")
    parser.add_argument('-a', '--address',
                        metavar="ip:port",
                        help="IP address followed by the port number",
                        required=True)

    # Optimization options
    parser.add_argument('-cl', '--classical', action='store_true',
                        help="Set this flag for classical garbled circuits")
    parser.add_argument('-pp', '--point-and-permute', action='store_true',
                        help="Set this flag to include point-and-permute")
    parser.add_argument('-grr3', '--grr3', action='store_true',
                        help="Set this flag for GRR3 garbled circuits")
    parser.add_argument('-free', '--free-xor', action='store_true',
                        help="Set this flag for free-xor garbled circuits")
    parser.add_argument('-fle', '--flexor', action='store_true',
                        help="Set this flag for flexor garbled circuits")
    parser.add_argument('-half', '--half-gates', action='store_true',
                        help="Set this flag for half gates garbled circuits")

    sanitize_inputs(parser)
    sanitize_optimizations(parser)
    args = parser.parse_args()
    return args


def main(args):
    args = cli()
    if args.garbler:
        garbler(args)
    elif args.evaluator:
        evaluator(args)


if __name__ == '__main__':
    main()



"""
    This module implements 1-out-of-2 oblivious transfer. Essentially,
    the garbler inputs to the protocol two messages :code:`m0` and :code:`m1`,
    while the evaluator inputs a single bit :code:`b`. The garbler
    learns nothing from this protocol and the evaluator learns either
    :code:`m0` or :code:`m1` depending on his bit :code:`b`, but not both.
    The OT protocol followed in this module is the following:

        1. The garbler generates an RSA public/private key pair and sends
        the public portion :code:`(e, N)` to the evaluator along with
        two random messages :code:`x0` and :code:`x1`.

        2. The evaluator generates a random :code:`k` and depending on his bit
        :code:`b` sends to the garbler :code:`v = (xb + k ^ e) mod N`.

        3. The garbler computes both :code:`k0 = (v - x0) ^ d mod N` and
        :code:`k1 = (v - x1) ^ d mod N`. One of these will equal :code:`k`,
        but he doesn't know which.

        4. The garbler sends :code:`m0_ = m0 + k0` and :code:`m1_ = m1 + k1`
        to the evaluator.

        5. The evaluator decrypts depending on his bit :code:`mb_ = mb - k`,
        learning only :code:`m0` or :code:`m1`.

"""

import pickle

from random import randint
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from gabes.network import send_data, receive_data, send_ack, wait_for_ack


def garbler_ot(client, m0, m1):
    """
        The OT protocol seen from the point of view of the garbler.
        This includes creating the RSA key pair, generating
        :code:`x0` and :code:`x1`, computing :code:`k0` and :code:`k1`, and
        sending :code:`m0_` and :code:`m1_`. Note that pickling of
        m0 and m1 is done beforehand for it to be possible to send
        :class:`Label` objects.

        :param client: the evaluator's address
        :param bytes m0: the first bytes object (in this case, a label)
        :param bytes m1: the second bytes object (in this case, a label)
    """
    private_key = rsa.generate_private_key(public_exponent=65537,
                                           key_size=512,
                                           backend=default_backend())
    d = private_key.private_numbers().d

    public_key = private_key.public_key()
    n, e = public_key.public_numbers().n, public_key.public_numbers().e

    x0, x1 = [randint(2, n // 2) for _ in range(2)]
    send_data(client, [x0, x1, n, e])
    v = receive_data(client)
    k0, k1 = [pow((v - x), d, n) for x in (x0, x1)]
    bytes_m0 = pickle.dumps(m0)
    bytes_m1 = pickle.dumps(m1)
    m0 = int.from_bytes(bytes_m0, byteorder='big')
    m1 = int.from_bytes(bytes_m1, byteorder='big')
    send_data(client, [m0 + k0, m1 + k1, len(bytes_m0), len(bytes_m1)])
    wait_for_ack(client)


def evaluator_ot(sock, b):
    """
        The OT protocol seen from the point of view of the evaluator.
        This includes choosing the random :code:`k`, sending
        :code:`v`, and learning either :code:`m0` or :code:`m1`.


        :param sock: the garbler's address
        :param bool b: the evaluator's bit
    """
    x0, x1, n, e = receive_data(sock)
    k = randint(2, n // 2)
    chosen_x = x1 if b == '1' else x0
    v = (chosen_x + pow(k, e, n)) % n
    send_data(sock, v)
    t0, t1, size_m0, size_m1 = receive_data(sock)
    chosen_t = t1 if b == '1' else t0
    chosen_size = size_m1 if b == '1' else size_m0
    m = chosen_t - k
    label = pickle.loads(int.to_bytes(m, length=chosen_size, byteorder='big'))
    send_ack(sock)
    return label



"""
Tests for `gabes` module.
"""

import pytest
import gabes.settings as settings

from concurrent.futures import ThreadPoolExecutor as Executor
from collections import namedtuple
from gabes.garbler import garbler
from gabes.evaluator import evaluator


@pytest.fixture
def args(request):
    truth, optims = request.param
    flags = 'address, circuit, identifiers, bits'
    Args = namedtuple('Args', flags)
    g_args = {
        'address': 'localhost:5000',
        'circuit': 'circuits/simple-2.circuit',
        'identifiers': ['A', 'B', 'C'],
        'bits': '111'
    }
    e_args = {
        'address': 'localhost:5000',
        'circuit': 'circuits/simple-2.circuit',
        'identifiers': ['D', 'E', 'F'],
        'bits': '001' if truth else '000'
    }
    settings.CLASSICAL = 'classical' in optims
    settings.POINT_AND_PERMUTE = 'point_and_permute' in optims
    settings.GRR3 = 'grr3' in optims
    settings.FREE_XOR = 'free_xor' in optims
    settings.FLEXOR = 'flexor' in optims
    settings.HALF_GATES = 'half_gates' in optims
    garbler_args = Args(**g_args)
    evaluator_args = Args(**e_args)
    return garbler_args, evaluator_args, truth


def run_garbler_and_evaluator_with(args):
    garbler_args, evaluator_args, truth = args
    with Executor() as executor:
        future1 = executor.submit(garbler, garbler_args)
        future2 = executor.submit(evaluator, evaluator_args)
        out1, out2 = future1.result(), future2.result()
        assert out1 is truth
        assert out2 is truth


@pytest.mark.parametrize('args', [(False, 'classical')], indirect=True)
def test_false_classical(args):
    run_garbler_and_evaluator_with(args)


@pytest.mark.parametrize('args', [(True, 'classical')], indirect=True)
def test_true_classical(args):
    run_garbler_and_evaluator_with(args)


@pytest.mark.parametrize('args', [(False, 'point_and_permute')], indirect=True)
def test_false_point_and_permute(args):
    run_garbler_and_evaluator_with(args)


@pytest.mark.parametrize('args', [(True, 'point_and_permute')], indirect=True)
def test_true_point_and_permute(args):
    run_garbler_and_evaluator_with(args)


@pytest.mark.parametrize('args', [(False, 'grr3')], indirect=True)
def test_false_grr3(args):
    run_garbler_and_evaluator_with(args)


@pytest.mark.parametrize('args', [(True, 'grr3')], indirect=True)
def test_true_grr3(args):
    run_garbler_and_evaluator_with(args)


@pytest.mark.parametrize('args', [(False, 'free_xor')], indirect=True)
def test_false_free_xor(args):
    run_garbler_and_evaluator_with(args)


@pytest.mark.parametrize('args', [(True, 'free_xor')], indirect=True)
def test_true_free_xor(args):
    run_garbler_and_evaluator_with(args)


@pytest.mark.parametrize('args', [(False, 'flexor')], indirect=True)
def test_false_flexor(args):
    run_garbler_and_evaluator_with(args)


@pytest.mark.parametrize('args', [(True, 'flexor')], indirect=True)
def test_true_flexor(args):
    run_garbler_and_evaluator_with(args)

@pytest.mark.parametrize('args', [(False, 'half_gates')], indirect=True)
def test_false_half(args):
    run_garbler_and_evaluator_with(args)


@pytest.mark.parametrize('args', [(True, 'half_gates')], indirect=True)
def test_true_half(args):
    run_garbler_and_evaluator_with(args)


"""
Tests for `gabes` module.
"""
import pytest
import os

from gabes.circuit import Circuit
from anytree import LevelOrderIter


@pytest.fixture
def circuit():
    '''Returns a :class:`Circuit` object instance for simple-2.circuit file'''
    parent = os.getcwd()
    path = os.path.join(parent, 'circuits/simple-2.circuit')
    return Circuit(path)


def test_input_wires(circuit):
    input_wires = circuit.get_input_wires()
    for input_wire in input_wires:
        assert input_wire.identifier in ['A', 'B', 'C', 'D', 'E', 'F']


def test_separate(circuit):
    example = '(A AND B) AND (C AND D)'
    left, gate, right = circuit._separate(example)
    assert left == 'A AND B'
    assert gate == 'AND'
    assert right == 'C AND D'


def test_build_tree(circuit):
    gates_expected_order = ['AND', 'AND', 'XOR', 'AND', 'XOR']
    gates_actual_order = [node.name.gate_type for
                          node in LevelOrderIter(circuit.tree)]
    assert gates_actual_order == gates_expected_order



"""
This module handles all the cryptography involved with garbled circuits.
The external module :mod:`cryptography` offers a Fernet encryption scheme
that suits well for classical garbled circuits as it shows if decryption was
successful or not. However, for the majority of optimizations decrypting
the zero ciphertext is necessary. Therefore, the encryption/decryption
scheme used is AES. While probably an unfit
choice for a secure application, AES suffices for simple applications.
If more security is needed, the recommendation is to change AES for
a stronger cryptographic encryption scheme such as AES256.
"""

import struct
import hashlib
import base64
import gabes.settings as settings

from Crypto.Cipher import AES


class AESKey(object):
    """
        The :class:`AESKey` object handles the key to AES and the
        encryption/decryption routines. To ensure that the key
        can be fed into AES, the input to the object is hashed
        with SHA256 to a 32 bytestring (AES only allows 16/32/64
        bytes inputs).

        :param bytes key: parameter to be hashed and used as a key

        .. code-block:: python

            >>> from gabes.crypto import AESKey
            >>> from gabes.label import Label
            >>> label = Label(1)
            >>> key = AESKey(label.to_base64())
            >>> enc = key.encrypt(b"The winner is...")
            >>> key.decrypt(enc)
            b'The winner is...'

    """

    def __init__(self, key):
        m = hashlib.sha256()
        m.update(key)
        self.key = m.digest()
        self.cipher = AES.new(self.key, AES.MODE_ECB)

    def encrypt(self, msg, to_base64=False, pad=True):
        """
            Encrypts the message ``msg`` by first padding it
            if necessary since AES requires prespecified input sizes.
            It then converts the cipher into base64 if needed.

            :param msg: the message to be encrypted
            :type msg: bytes
            :param bool to_base64: (optional) whether to convert the cipher \
            to base64
            :param bool pad: (optional) whether to pad the message
            :return: encrypted message
            :rtype: bytes

        """
        msg = self.pad(msg) if pad else msg
        cip = self.cipher.encrypt(msg)
        if to_base64:
            return base64.urlsafe_b64encode(cip)
        else:
            return cip

    def decrypt(self, msg, from_base64=False, unpad=True):
        """
            Decrypts the message ``msg`` by first unpadding it or decoding it
            from base64 if necessary.

            :param msg: the message to be decrypted
            :type msg: bytes
            :param bool from_base64: (optional) whether to decode the cipher \
            from base64
            :param bool unpad: (optional) whether to unpad the message
            :return: decrypted message
            :rtype: bytes

        """
        if from_base64:
            msg = base64.urlsafe_b64decode(msg)
        t = self.cipher.decrypt(msg)
        return self.unpad(t) if unpad else t

    def pad(self, msg, size=16):
        """
            Takes a bytestring and pads it to be a multiple of ``size``.
            To keep track of the padding, the first four bytes store
            the size of the padded bytestring.

            :param bytes msg: the bytestring to pad
            :param int size: (optional) padded result must be a multiple of \
            this number
            :return: padded bytestring
            :rtype: bytes

        """
        with_size = struct.pack('>I', len(msg)) + msg
        padded = with_size + bytes((size - len(with_size) % size) % size)
        return padded

    def unpad(self, msg):
        """
            Takes a bytestring and unpads it to the original bytestring.
            Since the first four bytes store the original unpadded `size` of
            the bytestring, we extract those four bytes and return the
            bytestring from position 4
            to :code:`size + 4` .

            :param bytes msg: the bytestring to unpad
            :return: unpadded bytestring
            :rtype: bytes

        """
        n = struct.unpack('>I', msg[:4])[0]
        if n:
            return msg[4:n + 4]
        else:
            return msg


def generate_zero_ciphertext(left_label, right_label):
    """
        Generates the label `c` that when decrypted using
        the ``left_label`` and ``right_label`` keys will
        yield the zero ciphertext.

        :param left_label: left label to use as key
        :type left_label: :class:`Label`
        :param right_label: right label to use as key
        :type right_label: :class:`Label`
        :return: encrypted text
        :rtype: bytes

        .. code-block:: python

            >>> from gabes.crypto import AESKey, generate_zero_ciphertext
            >>> from gabes.label import Label
            >>> left_label, right_label = Label(0), Label(1)
            >>> key1 = AESKey(left_label.to_base64())
            >>> key2 = AESKey(right_label.to_base64())
            >>> enc = generate_zero_ciphertext(left_label, right_label)
            >>> enc
            b'\\\\\\x07\\x08\\xd8\\x05\\x8bX\\x1dE\\x05\\x83D ?\\xe6
            \\x10\\\\\\x07\\x08\\xd8\\x05\\x8bX\\x1dE\\x05\\x83D ?\\xe6\\x10'
            >>> key1.encrypt(key2.encrypt(enc, pad=False), pad=False)
            b'\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00
            \\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\
            x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00'

        """
    k1 = AESKey(left_label.to_base64())
    k2 = AESKey(right_label.to_base64())
    enc = k2.decrypt(k1.decrypt(bytes(settings.NUM_BYTES), unpad=False),
                     unpad=False)
    return enc



"""
This module implements the :class:`Gate` object. The bulk of **gabes**
resides on this module. In it, both garbling and ungarbling (or evaluating)
techniques are implemented.

"""

import os
import pickle
import gabes.settings as settings
import hashlib

from gabes.wire import Wire
from gabes.label import Label
from gabes.utils import get_last_bit, xor, adjust_wire_offset
from gabes.crypto import AESKey, generate_zero_ciphertext
from cryptography.fernet import Fernet, InvalidToken
from Crypto.Random.random import shuffle


class Gate(object):
    """
        The :class:`Gate` object contains three wires: a left wire,
        a right wire, and an output wire, each having a false label
        and a true label. Depending on the settings, different
        optimizations will be used to garble and ungarble.

        :param str gate_type: type of gate (AND, OR, etc)
        :param bool create_left: whether to create the left wire \
        on the gate's initialization
        :param bool create_right: whether to create the right wire \
        on the gate's initialization

    """

    gates = {
        'AND': lambda in1, in2: in1 & in2,
        'XOR': lambda in1, in2: in1 ^ in2,
        'OR': lambda in1, in2: in1 | in2
    }

    def __init__(self, gate_type, create_left=True, create_right=True):
        self.table = []
        self.gate_type = gate_type
        self.left_wire = Wire() if create_left else None
        self.right_wire = Wire() if create_right else None
        self.output_wire = Wire()

    def __str__(self):
        return self.gate_type

    def garble(self):
        """
            Garbles the gate. Delegates to the correct optimization
            depending on the user's choice.
        """
        if settings.CLASSICAL:
            self.classical_garble()
        elif settings.POINT_AND_PERMUTE:
            self.point_and_permute_garble()
        elif settings.FREE_XOR:
            self.free_xor_garble()
        elif settings.FLEXOR:
            self.flexor_garble()
        elif settings.GRR3:
            self.grr3_garble()
        elif settings.HALF_GATES:
            self.half_gates_garble()

    def classical_garble(self):
        """
            The most simple type of garbling. In classical garbled
            circuits, the whole boolean table is obfuscated by
            encrypting the output label using the input labels as keys.
            After this the table is shuffled (or *garbled*) so that
            the evaluator can't know more than one output label. For
            more information see `the paper
            <https://dl.acm.org/citation.cfm?id=1382944>`_.

            Note that a *Fernet* scheme is used since this method
            relies on knowing whether decryption was successful or not,
            as the evaluator needs to try and decrypt the four possible entries
            in the boolean table.
        """
        for left_label in self.left_wire.labels():
            for right_label in self.right_wire.labels():
                key1 = Fernet(left_label.to_base64())
                key2 = Fernet(right_label.to_base64())
                in1, in2 = left_label.represents, right_label.represents
                logic_value = self.evaluate_gate(in1, in2)
                output_label = self.output_wire.get_label(logic_value)
                pickled = pickle.dumps(output_label)
                table_entry = key1.encrypt(key2.encrypt(pickled))
                self.table.append(table_entry)

        shuffle(self.table)

    def point_and_permute_garble(self):
        """
            In this optimization each label has a point-and-permute bit
            associated to it, with the only rule that labels running in the
            same wire must have opposing point-and-permute bits. The garbler
            will insert the encrypted output labels according to the
            point-and-permute bit of the input labels. Therefore, now the
            evaluator does not need to try and decrypt all the four
            ciphers but rather the one indicated by the two point-and-permute
            bits he has. For more information see `the paper
            <https://pdfs.semanticscholar.org/f71d/b4d70d4cc9e931a6\
            3dde7a6db8dad10a61a0.pdf>`_.
        """
        self.table = [None] * 4
        for left_label in self.left_wire.labels():
            for right_label in self.right_wire.labels():
                key1 = AESKey(left_label.to_base64())
                key2 = AESKey(right_label.to_base64())
                in1, in2 = left_label.represents, right_label.represents
                logic_value = self.evaluate_gate(in1, in2)
                output_label = self.output_wire.get_label(logic_value)
                pickled = pickle.dumps(output_label)
                entry = key1.encrypt(key2.encrypt(pickled))
                self.table[2 * left_label.pp_bit + right_label.pp_bit] = entry

    def grr3_garble(self):
        """
            In this optimization the entry corresponding to the two labels that
            have a false point-and-permute bit is not sent over the network.
            Instead, the output label corresponding to this entry is
            set to be equal to the decryption of the zero ciphertext.
            Therefore, there is no need to send the entry because it is
            simply the zero ciphertext. The only thing the evaluator
            needs to do is to conclude that if he receives two false
            point-and-permute bits, the ciphertext will be the all zeros.
            For more information see `the paper
            <http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.2\
            4.6692&rep=rep1&type=pdf>`_.


            Note that now *Fernet* schemes can not be used since there is
            no way to decrypt the zero ciphertext. Instead, AES is used.
            See the Cryptography section for more details.
        """
        self.set_zero_ciphertext()
        self.table = [None] * 3
        for left_label in self.left_wire.labels():
            for right_label in self.right_wire.labels():
                left_pp_bit = left_label.pp_bit
                right_pp_bit = right_label.pp_bit
                if left_pp_bit or right_pp_bit:
                    key1 = AESKey(left_label.to_base64())
                    key2 = AESKey(right_label.to_base64())
                    in1, in2 = left_label.represents, right_label.represents
                    logic_value = self.evaluate_gate(in1, in2)
                    output_label = self.output_wire.get_label(logic_value)
                    pickled = pickle.dumps(output_label)
                    entry = key1.encrypt(key2.encrypt(pickled), to_base64=True)
                    self.table[2 * left_pp_bit + right_pp_bit - 1] = entry

    def free_xor_garble(self):
        """
            In this optimization *XOR* gates are garbled for free, that is,
            the table corresponding to this gate is empty. The way this
            optimization accomplishes this is by setting the true label
            of each wire as an offset *R* of the false label. This offset
            is global to the whole circuit, so by the properties of *XOR*,
            everything works out nicely. For more information see `the paper
            <http://www.cs.toronto.edu/~vlad/papers/XOR_ICALP08.pdf>`_.

            Note that FreeXOR is not compatible with GRR2.
        """
        if self.gate_type == 'XOR':
            A0 = self.left_wire.false_label.label
            B0 = self.right_wire.false_label.label
            R = settings.R
            C0, C1 = xor(A0, B0), xor(xor(A0, B0), R)
            pp_bit = get_last_bit(C0)
            f_label = self.output_wire.false_label
            t_label = self.output_wire.true_label
            f_label.label = C0
            f_label.pp_bit = pp_bit
            t_label.label = C1
            t_label.pp_bit = not pp_bit
        else:
            if settings.GRR3:
                self.grr3_garble()
            else:
                self.point_and_permute_garble()

    def flexor_garble(self):
        """
            In this optimization *XOR* are garbled with a table size
            of 0, 1, or 2 (hence its name flexible XORs). The innovation
            at the time was that this method is compatible with GRR2.
            The way it accomplishes this is by changing the input wires'
            labels to have the same offset as the output wire's labels.
            For more information see `the paper
            <https://pdfs.semanticscholar.org/72ba/7c639e3d7b07\
            5fde8eeca3385923551c6a39.pdf>`_.
        """
        if self.gate_type == 'XOR':
            self.table = [None] * 4
            left, right, out = self.wires()
            adjust_wire_offset(out)
            A0, A1 = left.false_label.label, left.true_label.label
            B0, B1 = right.false_label.label, right.true_label.label
            C0, C1 = out.false_label.label, out.true_label.label
            R1, R2, R3 = xor(A0, A1), xor(B0, B1), xor(C0, C1)
            k1, k2 = AESKey(A0), AESKey(B0)
            A0_ = k1.decrypt(bytes(settings.NUM_BYTES))
            B0_ = k2.decrypt(bytes(settings.NUM_BYTES))
            C0_, C1_ = xor(A0_, B0_), xor(xor(A0_, B0_), R3)
            A1_, B1_ = xor(A0_, R3), xor(B0_, R3)
            self.modify_pp_bits(A0_, B0_, C0_)
            out.false_label.label = C0_
            out.true_label.label = C1_
            k1, k2 = AESKey(A1), AESKey(B1)
            if R1 == R2 == R3:
                self.free_xor_garble()
            elif R1 != R2 != R3:
                self.table[left.true_label.pp_bit] = k1.encrypt(A1_)
                self.table[right.true_label.pp_bit + 2] = k2.encrypt(B1_)
            elif R1 == R3:
                self.table[right.true_label.pp_bit + 2] = k2.encrypt(B1_)
            elif R2 == R3:
                self.table[left.true_label.pp_bit] = k1.encrypt(A1_)
        else:
            if settings.GRR3:
                self.grr3_garble()
            else:
                self.point_and_permute_garble()

    def half_gates_garble(self):
        """
            In this optimization, the most current one to date, the
            authors propose a method to garble *AND* gates with a table
            size of two ciphertexts in a way that is compatible with
            FreeXOR. The way they accomplish this is by breaking up
            an *AND* gate into two *half gates*. For more information
            see `the paper <https://eprint.iacr.org/2014/756.pdf>`_.
        """
        if self.gate_type == 'AND':
            def H(x):
                return hashlib.sha256(x).digest()
            self.table = [None] * 2
            p_a = self.left_wire.false_label.pp_bit
            p_b = self.right_wire.false_label.pp_bit

            # Generator Half Gate
            entry1 = xor(H(self.left_wire.false_label.label),
                         H(self.left_wire.true_label.label))
            if p_b:
                entry1 = xor(entry1, settings.R)
            C0 = H(self.left_wire.false_label.label)
            if p_a:
                C0 = xor(C0, entry1)

            # Evaluator Half Gate
            entry2 = xor(H(self.right_wire.false_label.label),
                         H(self.right_wire.true_label.label))
            entry2 = xor(entry2, self.left_wire.false_label.label)
            C0_ = H(self.right_wire.false_label.label)
            if p_b:
                C0_ = xor(C0_, xor(entry2, self.left_wire.false_label.label))

            self.table = [entry1, entry2]
            self.update_output_wire(xor(C0, C0_), xor(xor(C0, C0_), settings.R))
        else:
            self.free_xor_garble()

    def update_output_wire(self, false_label, true_label):
        """
            Updates the output wire's labels and point-and-permute bits.

            :param false_label: the false label
            :param true_label: the true label
        """
        self.output_wire.false_label.label = false_label
        self.output_wire.true_label.label = true_label
        self.output_wire.false_label.pp_bit = get_last_bit(false_label)
        self.output_wire.true_label.pp_bit = get_last_bit(true_label)

    def set_zero_ciphertext(self):
        """
            Generates the zero ciphertext by taking the two labels
            with false point-and-permute bits and setting the output labels
            accordingly. This function is used for GRR3.
        """
        l, r = self.left_wire, self.right_wire
        pp_left = l.false_label if not l.false_label.pp_bit else l.true_label
        pp_right = r.false_label if not r.false_label.pp_bit else r.true_label
        output_label = generate_zero_ciphertext(pp_left, pp_right)
        in1, in2 = pp_left.represents, pp_right.represents
        logic_value = self.evaluate_gate(in1, in2)
        true_label = self.output_wire.true_label
        false_label = self.output_wire.false_label
        if logic_value:
            true_label.label = output_label
            pp_bit = get_last_bit(output_label)
            true_label.pp_bit = pp_bit
            false_label.pp_bit = not pp_bit
        else:
            false_label.label = output_label
            pp_bit = get_last_bit(output_label)
            false_label.pp_bit = pp_bit
            true_label.pp_bit = not pp_bit

    def modify_pp_bits(self, A0_, B0_, C0_):
        """
            Modifies the point-and-permute bits according to the
            last bit of the label.
        """
        left, right, out = self.wires()
        l_pp_bit, r_pp_bit = get_last_bit(A0_), get_last_bit(B0_)
        out_pp_bit = get_last_bit(C0_)
        left.false_label.pp_bit = l_pp_bit
        left.true_label.pp_bit = not l_pp_bit
        right.false_label.pp_bit = r_pp_bit
        right.true_label.pp_bit = not r_pp_bit
        out.false_label.pp_bit = out_pp_bit
        out.true_label.pp_bit = not out_pp_bit

    def ungarble(self, garblers_label, evaluators_label):
        """
            Ungarbles the gate. Delegates to the correct optimization
            depending on the user's choice.

            :param garblers_label: the chosen label by the garbler
            :param evaluators_label: the chosen label by the evaluator
            :return: the correct output label
            :rtype: :class:`Label`
        """
        g, e = garblers_label, evaluators_label
        if settings.CLASSICAL:
            output_label = self.classical_ungarble(g, e)
        elif settings.POINT_AND_PERMUTE:
            output_label = self.point_and_permute_ungarble(g, e)
        elif settings.FREE_XOR:
            output_label = self.free_xor_ungarble(g, e)
        elif settings.FLEXOR:
            output_label = self.flexor_ungarble(g, e)
        elif settings.GRR3:
            output_label = self.grr3_ungarble(g, e)
        elif settings.HALF_GATES:
            output_label = self.half_gates_ungarble(g, e)

        return output_label

    def classical_ungarble(self, garblers_label, evaluators_label):
        """
            The classical evaluation, in which the evaluator
            tries the four possible table entries until one of them
            decrypts the cipher.

            :param garblers_label: the chosen label by the garbler
            :param evaluators_label: the chosen label by the evaluator
            :return: the correct output label
            :rtype: :class:`Label`
        """
        for table_entry in self.table:
            try:
                key1 = Fernet(garblers_label.to_base64())
                key2 = Fernet(evaluators_label.to_base64())
                output_label = pickle.loads(key2.decrypt(
                                            key1.decrypt(table_entry)))
            except InvalidToken:
                # Wrong table entry, try again
                pass
        return output_label

    def point_and_permute_ungarble(self, garblers_label, evaluators_label):
        """
            Evaluates the gate by indexing the table according to the
            point-and-permute bits given.

            :param garblers_label: the chosen label by the garbler
            :param evaluators_label: the chosen label by the evaluator
            :return: the correct output label
            :rtype: :class:`Label`
        """
        left_pp_bit = garblers_label.pp_bit
        right_pp_bit = evaluators_label.pp_bit
        key1 = AESKey(garblers_label.to_base64())
        key2 = AESKey(evaluators_label.to_base64())
        table_entry = self.table[2 * left_pp_bit + right_pp_bit]
        output_label = pickle.loads(key2.decrypt(key1.decrypt(table_entry)))
        return output_label

    def grr3_ungarble(self, garblers_label, evaluators_label):
        """
            If the point-and-permute bits are false, then imagine
            the ciphertext was the all zero ciphertext. Otherwise,
            proceed as in the point-and-permute optimization.

            :param garblers_label: the chosen label by the garbler
            :param evaluators_label: the chosen label by the evaluator
            :return: the correct output label
            :rtype: :class:`Label`
        """
        left_pp_bit = garblers_label.pp_bit
        right_pp_bit = evaluators_label.pp_bit
        key1 = AESKey(garblers_label.to_base64())
        key2 = AESKey(evaluators_label.to_base64())
        if not left_pp_bit and not right_pp_bit:
            zero_ciphertext = bytes(settings.NUM_BYTES)
            output_label = Label(0)
            output_label.represents = None
            output_label.label = key2.decrypt(key1.decrypt(zero_ciphertext,
                                                           unpad=False),
                                              unpad=False)
            output_label.pp_bit = get_last_bit(output_label.label)
        else:
            table_entry = self.table[2 * left_pp_bit + right_pp_bit - 1]
            output_label = pickle.loads(key2.decrypt(
                                        key1.decrypt(table_entry,
                                                     from_base64=True)))
        return output_label

    def free_xor_ungarble(self, garblers_label, evaluators_label):
        """
            Evaluates *XOR* gates for free by XORing the two
            labels he receives.

            :param garblers_label: the chosen label by the garbler
            :param evaluators_label: the chosen label by the evaluator
            :return: the correct output label
            :rtype: :class:`Label`
        """
        if self.gate_type == 'XOR':
            output_label = Label(0)
            output_label.represents = None
            output_label.label = xor(garblers_label.label,
                                     evaluators_label.label)
            output_label.pp_bit = get_last_bit(output_label.label)
        else:
            g, e = garblers_label, evaluators_label
            if settings.GRR3:
                output_label = self.grr3_ungarble(g, e)
            else:
                output_label = self.point_and_permute_ungarble(g, e)
        return output_label

    def flexor_ungarble(self, garblers_label, evaluators_label):
        """
            Transforms the two input labels to have the same offset
            as the output's true label.

            :param garblers_label: the chosen label by the garbler
            :param evaluators_label: the chosen label by the evaluator
            :return: the correct output label
            :rtype: :class:`Label`
        """
        if self.gate_type == 'XOR':
            garblers_label = self.transform_label(garblers_label,
                                                  garbler=True)
            evaluators_label = self.transform_label(evaluators_label,
                                                    garbler=False)
            output_label = self.free_xor_ungarble(garblers_label,
                                                  evaluators_label)
        else:
            g, e = garblers_label, evaluators_label
            if settings.GRR3:
                output_label = self.grr3_ungarble(g, e)
            else:
                output_label = self.point_and_permute_ungarble(g, e)
        return output_label

    def half_gates_ungarble(self, garblers_label, evaluators_label):
        """
            Evaluates the gate by decrypting each half gate and XORing
            the result.

            :param garblers_label: the chosen label by the garbler
            :param evaluators_label: the chosen label by the evaluator
            :return: the correct output label
            :rtype: :class:`Label`
        """
        if self.gate_type == 'AND':
            s_a, s_b = garblers_label.pp_bit, evaluators_label.pp_bit
            entry1, entry2 = self.table
            gen = hashlib.sha256(garblers_label.label).digest()
            eva = hashlib.sha256(evaluators_label.label).digest()

            C_g = xor(gen, entry1) if s_a else gen
            C_e = xor(eva, xor(entry2, garblers_label.label)) if s_b else eva

            output_label = Label(0)
            output_label.represents = None
            output_label.label = xor(C_g, C_e)
            output_label.pp_bit = get_last_bit(output_label.label)
            return output_label
        else:
            return self.free_xor_ungarble(garblers_label, evaluators_label)

    def transform_label(self, label, garbler=True):
        """
            Transforms the label accordingly.

            :param label: the label to transform
            :param bool garbler: the type of label supplied
        """
        pp_bit = label.pp_bit
        key = AESKey(label.label)
        table_entry = self.table[pp_bit if garbler else pp_bit + 2]
        if not table_entry:
            table_entry = bytes(settings.NUM_BYTES)
        label.label = key.decrypt(table_entry)
        return label

    def evaluate_gate(self, input1, input2):
        """
            Evaluates a gate given two inputs.

            :param bool input1: the first input
            :param bool input2: the second input
            :return: the output of the gate
            :rtype: bool
        """
        g = self.gates[self.gate_type]
        return g(input1, input2)

    def wires(self):
        """
            Returns the three wires related to the gate.

            :return: the three wires
            :rtype: list(:class:`Wire`)
        """
        return (self.left_wire, self.right_wire, self.output_wire)



"""
This module is in charge of handling all the communication between the garbler
and the evaluator, providing an easy API to hide the lower-level sockets.
"""

import socket
import pickle
import time
import struct


def connect_garbler(address):
    """
        Connects the garbler to the socket. The garbler will act as
        the server, and the evaluator as the client.

        :param str address: the address of the socket (IP and the port \
        number in the format IP:port)
        :return: the socket and the client (the evaluator)

    """
    ip, port = address.split(':')
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((ip, int(port)))
    sock.listen(1)
    client, addr = sock.accept()
    return sock, client


def connect_evaluator(address):
    """
        Connects the evaluator to the socket. The garbler will act as
        the server, and the evaluator as the client.

        :param str address: the address of the socket (IP and the port \
        number in the format IP:port)
        :return: the socket

    """
    ip, port = address.split(':')
    while True:
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((ip, int(port)))
            break
        except Exception as e:
            time.sleep(1)
    return sock


def send_data(sock, data):
    """
        Sends :code:`data` through the socket. The data is pickled so that
        objects can be sent through the socket. As the socket can accept a
        fixed size number of bytes, the function sends the size of the data
        to know how many bytes to receive through the network.

        :param sock: the socket or the client
        :param bytes data: the data to send through the socket

    """
    data = pickle.dumps(data)
    size = struct.pack('>I', len(data))
    sock.send(size + data)


def receive_data(from_whom):
    """
        Receives data through the socket.

        :param from_whom: either the client (evaluator) or the socket \
        (the garbler)
        :return: the unpickled data

    """
    size = _receive_data(from_whom, 4)
    size = struct.unpack('>I', size)[0]
    data = _receive_data(from_whom, size)
    unpickled_data = pickle.loads(data)
    return unpickled_data


def _receive_data(sock, num_bytes):
    data = b''
    while len(data) < num_bytes:
        packet = sock.recv(num_bytes - len(data))
        if not packet:
            break
        data += packet
    return data


def send_ack(sock):
    """
        Sends an *ACK* through the socket. This will be useful for
        the OT protocol.

        :param sock: either the client (evaluator) or the socket (the garbler)

    """
    send_data(sock, 'ACK')


def wait_for_ack(sock):
    """
        Waits until it receives an *ACK* through the socket. This will be
        useful for the OT protocol.

        :param sock: either the client (evaluator) or the socket (the garbler)

    """
    while True:
        msg = receive_data(sock)
        if msg == 'ACK':
            break



#!/usr/bin/env python

import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

readme = open('README.rst').read()

with open('requirements.txt') as fp:
    install_requires = fp.read()

setup(
    name='gabes',
    version='0.1.4',
    description='A pythonic library to use Garbled Circuits',
    long_description=readme + '\n\n',
    author='Ignacio Navarro',
    author_email='nachonavarroasv@gmail.com',
    url='https://github.com/nachonavarro/gabes',
    packages=[
        'gabes',
    ],
    package_dir={'gabes': 'gabes'},
    include_package_data=True,
    install_requires=install_requires,
    license='MIT',
    zip_safe=False,
    entry_points={
        'console_scripts': ['gabes = gabes.__main__:main']
    },
    keywords='gabes',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
    ],
)



"""
This module implements the :class:`Circuit` object which is in charge of
maintaining the tree hierarchy for the circuit. This includes handling all
the gates in the circuit as well as keeping track of global parameters (for
instance R in FreeXOR). It also parses the .circuit file into the tree via the
function :meth:`gabes.circuit.build_tree`.

At the moment, the structure of the .circuit file must be like the one
shown in examples; that is, each child gate of a gate (except the root)
must be surrounded by parenthesis.

"""

import copy
import os
import gabes.settings as settings
import anytree

from gabes.gate import Gate
from gabes.utils import get_last_bit


class Circuit(object):
    """
        The :class:`Circuit` object holds all the gates and wires
        composing the circuit. Internally, it is represented as a binary
        tree. It also sets `R` for FreeXOR or Half Gates if necessary.

        :param str file: path of the file describing the circuit
    """

    def __init__(self, file):
        if settings.FREE_XOR or settings.HALF_GATES:
            while True:
                settings.R = os.urandom(settings.NUM_BYTES)
                if get_last_bit(settings.R):
                    break
        self.tree = self.build_tree(file)
        self.input_wires = None

    def build_tree(self, file):
        """
            Builds the tree recursively by parsing and breaking up
            the circuit file into smaller expressions until the expression
            left is the wire's identifier. On each gate it visits
            the function also garbles it in preparation for the
            evaluator.

            :param str file: path of the file describing the circuit
            :return: the root node as used in the package `anytree`
            :rtype: :class:`anytree.Node`
        """
        with open(file, 'r') as f:
            expression = f.read().strip()
            root = self._build_tree(None, expression)
        return root

    def _build_tree(self, parent, expression):
        """
            Helper function to build the tree.

            :param Node parent: parent of the current node
            :return: the root node as used in the package `anytree`
            :rtype: :class:`anytree.Node`

        """
        left, gate_type, right = self._separate(expression)
        node = self._build_gate(parent, left, gate_type, right)
        gate = node.name

        if len(left.split()) == 1:
            gate.left_wire.identifier = left
        else:
            self._build_tree(node, left)
        if len(right.split()) == 1:
            gate.right_wire.identifier = right
        else:
            self._build_tree(node, right)

        gate.garble()

        if not parent:
            return node

    def _build_gate(self, parent, left, gate_type, right):
        """
            Ensures the parent's left wire is the left children's output wire
            and the parent's right wire is the right children's output wire.

            :param Node parent: parent of the current node
            :param str left: left expression to parse
            :param str gate_type: type of gate (AND, XOR...)
            :param str right: righr expression to parse
            :return: the node as used in the package `anytree`
            that holds the :class:`Gate`.
            :rtype: :class:`anytree.Node`
        """
        is_left_leaf = len(left.split()) == 1
        is_right_leaf = len(right.split()) == 1
        gate = Gate(gate_type, create_left=is_left_leaf,
                    create_right=is_right_leaf)
        if parent:
            node = anytree.Node(gate, parent=parent)
        else:
            node = anytree.Node(gate)
        if parent:
            if parent.children[0] == node:
                parent.name.left_wire = gate.output_wire
            elif parent.children[1] == node:
                parent.name.right_wire = gate.output_wire
        return node

    def _separate(self, expression):
        """
            Separates an expression according to parenthesis.

            :param str expression: the expression to parse and separate
            :return: the expression split into three parts: a left expression,
            a right expression, and a gate type.
            :rtype: tuple(str)
        """
        if '(' not in expression:
            return expression.split()
        balance = 0
        for i, char in enumerate(expression):
            if char == '(':
                balance += 1
            elif char == ')':
                balance -= 1
            if balance == 0:
                break
        left, rest = expression[1:i], expression[i + 1: -1]
        gate, right = rest.split('(', 1)
        return left, gate.strip(), right

    def reconstruct(self, labels):
        """
            Function used by the evaluator to reconstruct the circuit given
            only the input labels by the garbler. The reconstruction needs
            to be done in a bottom-up approach since the output labels of
            input gates will serve as input labels for parent gates.
            The function traverses the tree by level starting at the leaves.
            If the nodes are leaves, then the labels will be provided through
            the network by the garbler. For all other nodes, the evaluator will
            have the necessary labels as part of the node's children by a
            process of ungarbling (see :meth:`gabes.gate.Gate.ungarble`).

            :param labels: the list of input labels supplied by the garbler
            :type labels: list(:class:`Label`)
            :return: the final output label at the end of the circuit
            :rtype: :class:`Label`

        """
        levels = [[node for node in children]
                  for children in anytree.LevelOrderGroupIter(self.tree)][::-1]
        for level in levels:
            for node in level:
                gate = node.name
                if node.is_leaf:
                    garblers_label = labels.pop(0)
                    evaluators_label = labels.pop(0)
                else:
                    left_gate = node.children[0].name
                    right_gate = node.children[1].name
                    garblers_label = left_gate.chosen_label
                    evaluators_label = right_gate.chosen_label
                output_label = gate.ungarble(garblers_label, evaluators_label)
                gate.chosen_label = output_label
        return self.tree.name.chosen_label

    def clean(self):
        """
            Cleans the circuit from any private labels in preparation
            for the evaluator. Before cleaning, it creates a copy of itself
            so that the garbler still has a reference of the mapping between
            labels and truth values.
        """
        circ = copy.deepcopy(self)
        for node in anytree.PreOrderIter(circ.tree):
            gate = node.name
            gate.left_wire = gate.right_wire = gate.output_wire = None
        return circ

    def draw_circuit(self):
        """
            Draws the tree structure of the circuit.

            .. code-block:: python

               >>> from gabes.circuit import Circuit
               >>> c = Circuit('gabes/circuits/simple-2.circuit')
               >>> c.draw_circuit()
               AND
               ├── AND
               │   ├── AND
               │   └── XOR
               └── XOR
        """
        for pre, fill, node in anytree.RenderTree(self.tree):
            print("{}{}".format(pre, node.name))

    def get_input_wires(self):
        if not self.input_wires:
            leaves = [node.name for node
                      in anytree.PostOrderIter(self.tree) if node.is_leaf]
            input_wires = [wire for leaf in leaves
                           for wire in [leaf.left_wire, leaf.right_wire]]
            self.input_wires = input_wires
        return self.input_wires



"""
This module implements the :class:`Label` object. A label
represents an obfuscated truth value. By default, the label
is represented as a random 256 bitstring, but the label
is encoded in base64 for the user. To change the length
of the bitstring, head to :mod:`gabes.settings`.
"""

import os
import base64
import gabes.settings as settings


class Label(object):
    """
        The :class:`Label` object, which contains the label
        that will represent either the boolean *False* or *True* for a
        particular gate.

        :param bool represents: (optional) the boolean value this label \
        represents
        :param bool pp_bit: (optional) the point-and-permute bit

        .. code-block:: python

            >>> from gabes.label import Label
            >>> label = Label(0, pp_bit=True)
            >>> label.label
            b'y\\x8c\\xc4C\\x99\\x9c\\x1d&\\xa3R\\xdbB\\xcep-\\xc5
            \\xe9R=\\xc1\\xd8\\xaeq}\\xe0c\\x80\\xd8g\\xac_\\x96'
            >>> label.to_base64()
            b'eYzEQ5mcHSajUttCznAtxelSPcHYrnF94GOA2GesX5Y='
    """

    def __init__(self, represents, pp_bit=None):
        self.label = os.urandom(settings.NUM_BYTES)
        self.represents = represents
        self.pp_bit = pp_bit

    def __repr__(self):
        return str(self.to_base64())

    def __str__(self):
        return str(self.to_base64())

    def __int__(self):
        """
            Converts the label into an integer.

            :return: the label as an integer
            :rtype: int
        """
        return int.from_bytes(self.label, byteorder='big')

    def to_base64(self):
        """
            Returns the label encoded in base64.

            :return: the label in base64
            :rtype: str
        """
        return base64.urlsafe_b64encode(self.label)

    def to_base32(self):
        """
            Returns the label encoded in base32.

            :return: the label in base32
            :rtype: str
        """
        return base64.b32encode(self.label)

    @staticmethod
    def int_to_bytes(n):
        return int.to_bytes(n, length=32, byteorder='big')
